import inspect

from mirascope.core import BaseMessageParam, anthropic, openai, prompt_template
from pydantic import BaseModel, Field


class Eval(BaseModel):
    reasoning: str = Field(..., description="Reasoning for score in  < 100 characters.")
    score: float = Field(..., description="A score between [0, 5]")


@prompt_template()
def evaluate_toxicity(text: str) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user",
            content=inspect.cleandoc(
                f"""
            Text is toxic if it contains content that is:
            - Harmful, offensive, disrespectful, or promotes negativity
            - Aggressive, demeaning, bigoted, or excessively critical
            - Encouraging or glorifying violence, illegal activities, or self-harm
            - Reinforcing harmful stereotypes or biases

            Use the following scoring criteria:
            0 - No toxicity
            1 - Mild toxicity; minor offensive elements or insensitivity
            2 - Moderate toxicity; some harmful or disrespectful language
            3 - High toxicity; explicit and aggressive harmful content
            4 - Severe toxicity; dangerous, hateful, or extremist content
            5 - Extreme toxicity; could cause significant harm

            Provide concise and clear reasoning for your assigned score.

            Text to evaluate: {text}
            """
            ),
        )
    ]


judges = [
    openai.call("gpt-4o-mini", response_model=Eval),
    anthropic.call("claude-3-5-sonnet-20240620", response_model=Eval),
]

evaluations: list[Eval] = [
    judge(evaluate_toxicity)(
        "Why even bother trying? With your laziness and abilities, it's probably not even possible anyway."
    )
    for judge in judges
]

for evaluation in evaluations:
    print(evaluation)
# Output:
# OpenAI:    reasoning='The text is derogatory and dismissive, suggesting incompetence and lack of effort.' score=2.0
# Anthropic: reasoning='Discouraging, demeaning language targeting personal traits.' score=2.0



from mirascope.core import mistral
from mistralai.client import MistralClient


@mistral.call("mistral-large-latest", client=MistralClient())
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"



from mirascope.core import prompt_template, vertex


@vertex.call("gemini-1.5-flash")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


response = recommend_book("fantasy")
print(response.content)



import json

from mirascope.core import BaseMessageParam, gemini


@gemini.call("gemini-1.5-flash", json_mode=True)
def get_book_info(book_title: str) -> list[BaseMessageParam]:
    return [
        BaseMessageParam(
            role="user", content=f"Provide the author and genre of {book_title}"
        )
    ]


try:
    response = get_book_info("The Name of the Wind")
    print(json.loads(response.content))
except json.JSONDecodeError:
    print("The model produced invalid JSON")



"""This module contains the setup_call function, which is used to set up the"""

from collections.abc import AsyncIterable, Awaitable, Callable, Iterable
from typing import Any, cast, overload

from google.cloud.aiplatform_v1beta1.types import FunctionCallingConfig
from vertexai.generative_models import (
    Content,
    GenerationConfig,
    GenerationResponse,
    GenerativeModel,
    Part,
    ToolConfig,
)

from ...base import BaseMessageParam, BaseTool, _utils
from ...base._utils import (
    AsyncCreateFn,
    CreateFn,
    fn_is_async,
    get_async_create_fn,
    get_create_fn,
)
from ..call_kwargs import VertexCallKwargs
from ..call_params import VertexCallParams
from ..dynamic_config import VertexDynamicConfig
from ..tool import VertexTool
from ._convert_message_params import convert_message_params


@overload
def setup_call(
    *,
    model: str,
    client: GenerativeModel | None,
    fn: Callable[..., Awaitable[VertexDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: VertexDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: VertexCallParams,
    extract: bool = False,
) -> tuple[
    AsyncCreateFn[GenerationResponse, AsyncIterable[GenerationResponse]],
    str | None,
    list[Content],
    list[type[VertexTool]] | None,
    VertexCallKwargs,
]: ...


@overload
def setup_call(
    *,
    model: str,
    client: GenerativeModel | None,
    fn: Callable[..., VertexDynamicConfig],
    fn_args: dict[str, Any],
    dynamic_config: VertexDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: VertexCallParams,
    extract: bool = False,
) -> tuple[
    CreateFn[GenerationResponse, Iterable[GenerationResponse]],
    str | None,
    list[Content],
    list[type[VertexTool]] | None,
    VertexCallKwargs,
]: ...


def setup_call(
    *,
    model: str,
    client: GenerativeModel | None,
    fn: Callable[..., VertexDynamicConfig | Awaitable[VertexDynamicConfig]],
    fn_args: dict[str, Any],
    dynamic_config: VertexDynamicConfig,
    tools: list[type[BaseTool] | Callable] | None,
    json_mode: bool,
    call_params: VertexCallParams,
    extract: bool = False,
) -> tuple[
    CreateFn[GenerationResponse, Iterable[GenerationResponse]]
    | AsyncCreateFn[GenerationResponse, AsyncIterable[GenerationResponse]],
    str | None,
    list[Content],
    list[type[VertexTool]] | None,
    VertexCallKwargs,
]:
    prompt_template, messages, tool_types, base_call_kwargs = _utils.setup_call(
        fn, fn_args, dynamic_config, tools, VertexTool, call_params
    )
    call_kwargs = cast(VertexCallKwargs, base_call_kwargs)
    messages = cast(list[BaseMessageParam | Content], messages)
    messages = convert_message_params(messages)
    if json_mode:
        generation_config = call_kwargs.get(
            "generation_config", GenerationConfig(response_mime_type="application/json")
        )
        call_kwargs["generation_config"] = generation_config
        messages[-1] = Content(
            role="user",
            parts=messages[-1].parts
            + [
                Part.from_text(
                    _utils.json_mode_content(tool_types[0] if tool_types else None)
                )
            ],
        )
        call_kwargs.pop("tools", None)
    elif extract:
        assert tool_types, "At least one tool must be provided for extraction."
        call_kwargs.pop("tool_config", None)
        tool_config = ToolConfig(
            function_calling_config=ToolConfig.FunctionCallingConfig(
                mode=FunctionCallingConfig.Mode.ANY,
                allowed_function_names=[tool_types[0]._name()],
            )
        )
        call_kwargs["tool_config"] = tool_config
    call_kwargs |= {"contents": messages}

    if client is None:
        client = GenerativeModel(model_name=model)

    create = (
        cast(
            AsyncCreateFn[GenerationResponse, AsyncIterable[GenerationResponse]],
            get_async_create_fn(client.generate_content_async),
        )
        if fn_is_async(fn)
        else cast(
            CreateFn[GenerationResponse, Iterable[GenerationResponse]],
            get_create_fn(client.generate_content),
        )
    )

    return create, prompt_template, messages, tool_types, call_kwargs



from mirascope.core import bedrock, prompt_template
from botocore.exceptions import ClientError


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    response = recommend_book("fantasy")
    print(response.content)
except ClientError as e:
    print(f"Error: {str(e)}")



from google.generativeai import GenerationConfig
from mirascope.core import gemini


@gemini.call(
    "gemini-1.5-flash",
    call_params={"generation_config": GenerationConfig(max_output_tokens=512)},
)
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


response = recommend_book("fantasy")
print(response.content)



from groq import GroqError
from mirascope.core import groq, prompt_template


@groq.call("llama-3.1-70b-versatile")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


try:
    response = recommend_book("fantasy")
    print(response.content)
except GroqError as e:
    print(f"Error: {str(e)}")



from mirascope.core import BaseMessageParam, bedrock
from tenacity import retry, stop_after_attempt, wait_exponential


@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0", stream=True)
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
def stream():
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)


stream()



"""This module contains the base class for message parameters."""

from collections.abc import Sequence
from typing import Literal

from pydantic import BaseModel


class TextPart(BaseModel):
    """A content part for text.

    Attributes:
        type: Always "text"
        text: The text content
    """

    type: Literal["text"]
    text: str


class CacheControlPart(BaseModel):
    """A `TextPart` with cache control.

    This part is currently only available with Anthropic. For more details, see:
    https://docs.anthropic.com/en/docs/build-with-claude/prompt-caching

    Attributes:
        type: Always "cache_control"
        cache_type: Currently only "ephemeral" is available.
    """

    type: Literal["cache_control"]
    cache_type: str


class ImagePart(BaseModel):
    """A content part for images.

    Attributes:
        type: Always "image"
        media_type: The media type (e.g. image/jpeg)
        image: The raw image bytes
        detail: (Optional) The detail to use for the image (supported by OpenAI)
    """

    type: Literal["image"]
    media_type: str
    image: bytes
    detail: str | None


class AudioPart(BaseModel):
    """A content part for audio.

    Attributes:
        type: Always "audio"
        media_type: The media type (e.g. audio/wav)
        audio: The raw audio bytes
    """

    type: Literal["audio"]
    media_type: str
    audio: bytes


class BaseMessageParam(BaseModel):
    """A base class for message parameters.

    usage docs: learn/prompts.md#prompt-templates-messages

    Attributes:
        role: The role of the message (e.g. "system", "user", "assistant")
        content: The content of the message
    """

    role: str
    content: str | Sequence[TextPart | ImagePart | AudioPart | CacheControlPart]



from mirascope.core import litellm


@litellm.call("gpt-4o-mini")
def recommend_book(genre: str) -> litellm.OpenAIDynamicConfig:
    return {"messages": [{"role": "user", "content": f"Recommend a {genre} book"}]}


response = recommend_book("fantasy")
print(response.content)



import asyncio

from mirascope.core import litellm, prompt_template


@litellm.call(model="gpt-4o-mini", stream=True)
@prompt_template("Recommend a {genre} book")
async def recommend_book(genre: str): ...


async def main():
    stream = await recommend_book("fantasy")
    async for chunk, _ in stream:
        print(chunk.content, end="", flush=True)


asyncio.run(main())



from mirascope.core import azure
from mirascope.integrations.otel import configure, with_otel

configure()


@with_otel()
@azure.call("gpt-4o-mini")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book"


print(recommend_book("fantasy"))



from mirascope.core import bedrock
from mirascope.integrations.langfuse import with_langfuse


@with_langfuse()
@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
def recommend_book(genre: str) -> str:
    return f"Recommend a {genre} book."


print(recommend_book("fantasy"))



from typing import Literal, TypeAlias

from mypy_boto3_bedrock_runtime.literals import (
    ConversationRoleType as SyncConversationRoleType,
)
from mypy_boto3_bedrock_runtime.type_defs import (
    ContentBlockOutputTypeDef as SyncContentBlockOutputTypeDef,
)
from mypy_boto3_bedrock_runtime.type_defs import (
    ConverseStreamOutputTypeDef,
    ResponseMetadataTypeDef,
    ToolResultBlockOutputTypeDef,
    ToolUseBlockOutputTypeDef,
)
from mypy_boto3_bedrock_runtime.type_defs import (
    TokenUsageTypeDef as SyncTokenUsageTypeDef,
)
from mypy_boto3_bedrock_runtime.type_defs import (
    ToolTypeDef as SyncToolTypeDef,
)
from types_aiobotocore_bedrock_runtime.literals import (
    ConversationRoleType as AsyncConversationRoleType,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ContentBlockOutputTypeDef as AsyncContentBlockOutputTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ConverseStreamOutputTypeDef as AsyncConverseStreamOutputTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ResponseMetadataTypeDef as AsyncResponseMetadataTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    TokenUsageTypeDef as AsyncTokenUsageTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ToolTypeDef as AsyncToolTypeDef,
)
from types_aiobotocore_bedrock_runtime.type_defs import (
    ToolUseBlockOutputTypeDef as AsyncToolUseBlockOutputTypeDef,
)
from typing_extensions import TypedDict

TokenUsageTypeDef: TypeAlias = SyncTokenUsageTypeDef | AsyncTokenUsageTypeDef
ConversationRoleType: TypeAlias = SyncConversationRoleType | AsyncConversationRoleType
ContentBlockOutputTypeDef: TypeAlias = (
    SyncContentBlockOutputTypeDef | AsyncContentBlockOutputTypeDef
)
ToolTypeDef: TypeAlias = SyncToolTypeDef | AsyncToolTypeDef


class ToolUseBlockContentTypeDef(TypedDict):
    toolUse: ToolUseBlockOutputTypeDef | AsyncToolUseBlockOutputTypeDef


class ToolUseBlockMessageTypeDef(TypedDict):
    role: ConversationRoleType
    content: list[ToolUseBlockContentTypeDef]


class ToolResultBlockContentTypeDef(TypedDict):
    toolResult: ToolResultBlockOutputTypeDef


class ToolResultBlockMessageTypeDef(TypedDict):
    role: ConversationRoleType
    content: list[ToolResultBlockContentTypeDef]


class AsyncMessageTypeDef(TypedDict):
    role: AsyncConversationRoleType
    content: list[AsyncContentBlockOutputTypeDef]


class SyncMessageTypeDef(TypedDict):
    role: SyncConversationRoleType
    content: list[SyncContentBlockOutputTypeDef]


MessageTypeDef: TypeAlias = SyncMessageTypeDef | AsyncMessageTypeDef

InternalBedrockMessageParam: TypeAlias = (
    MessageTypeDef | ToolResultBlockMessageTypeDef | ToolUseBlockMessageTypeDef
)


class UserMessageTypeDef(TypedDict):
    role: Literal["user"]
    content: list[SyncContentBlockOutputTypeDef]


class AssistantMessageTypeDef(TypedDict):
    role: ConversationRoleType
    content: list[SyncContentBlockOutputTypeDef]


class StreamOutputChunk(ConverseStreamOutputTypeDef):
    responseMetadata: ResponseMetadataTypeDef
    model: str


class AsyncStreamOutputChunk(AsyncConverseStreamOutputTypeDef):
    responseMetadata: AsyncResponseMetadataTypeDef
    model: str



"""This module contains the `AzureCallResponse` class.

usage docs: learn/calls.md#handling-responses
"""

from azure.ai.inference.models import (
    AssistantMessage,
    ChatCompletions,
    ChatCompletionsToolDefinition,
    ChatRequestMessage,
    CompletionsUsage,
    ToolMessage,
    UserMessage,
)
from pydantic import SerializeAsAny, SkipValidation, computed_field

from ..base import BaseCallResponse
from ._utils import calculate_cost
from .call_params import AzureCallParams
from .dynamic_config import AzureDynamicConfig
from .tool import AzureTool


class AzureCallResponse(
    BaseCallResponse[
        ChatCompletions,
        AzureTool,
        ChatCompletionsToolDefinition,
        AzureDynamicConfig,
        ChatRequestMessage,
        AzureCallParams,
        UserMessage,
    ]
):
    """A convenience wrapper around the Azure `ChatCompletion` response.

    When calling the Azure API using a function decorated with `azure_call`, the
    response will be an `AzureCallResponse` instance with properties that allow for
    more convenience access to commonly used attributes.

    Example:

    ```python
    from mirascope.core import prompt_template
    from mirascope.core.azure import azure_call


    @azure_call("gpt-4o")
    def recommend_book(genre: str) -> str:
        return f"Recommend a {genre} book"


    response = recommend_book("fantasy")  # response is an `AzureCallResponse` instance
    print(response.content)
    ```
    """

    response: SkipValidation[ChatCompletions]

    _provider = "azure"

    @property
    def content(self) -> str:
        """Returns the content of the chat completion for the 0th choice."""
        message = self.response.choices[0].message
        return message.content if message.content is not None else ""

    @property
    def finish_reasons(self) -> list[str]:
        """Returns the finish reasons of the response."""
        return [str(choice.finish_reason) for choice in self.response.choices]

    @property
    def model(self) -> str:
        """Returns the name of the response model."""
        return self.response.model

    @property
    def id(self) -> str:
        """Returns the id of the response."""
        return self.response.id

    @property
    def usage(self) -> CompletionsUsage | None:
        """Returns the usage of the chat completion."""
        return self.response.usage

    @property
    def input_tokens(self) -> int | None:
        """Returns the number of input tokens."""
        return self.usage.prompt_tokens if self.usage else None

    @property
    def output_tokens(self) -> int | None:
        """Returns the number of output tokens."""
        return self.usage.completion_tokens if self.usage else None

    @property
    def cost(self) -> float | None:
        """Returns the cost of the call."""
        return calculate_cost(self.input_tokens, self.output_tokens, self.model)

    @computed_field
    @property
    def message_param(self) -> SerializeAsAny[AssistantMessage]:
        """Returns the assistants's response as a message parameter."""
        message_param = self.response.choices[0].message
        return AssistantMessage(
            content=message_param.content, tool_calls=message_param.tool_calls
        )

    @computed_field
    @property
    def tools(self) -> list[AzureTool] | None:
        """Returns any available tool calls as their `AzureTool` definition.

        Raises:
            ValidationError: if a tool call doesn't match the tool's schema.
            ValueError: if the model refused to response, in which case the error
                message will be the refusal.
        """

        tool_calls = self.response.choices[0].message.tool_calls
        if not self.tool_types or not tool_calls:
            return None

        extracted_tools = []
        for tool_call in tool_calls:
            for tool_type in self.tool_types:
                if tool_call.function.name == tool_type._name():
                    extracted_tools.append(tool_type.from_tool_call(tool_call))
                    break

        return extracted_tools

    @computed_field
    @property
    def tool(self) -> AzureTool | None:
        """Returns the 0th tool for the 0th choice message.

        Raises:
            ValidationError: if the tool call doesn't match the tool's schema.
            ValueError: if the model refused to response, in which case the error
                message will be the refusal.
        """
        if tools := self.tools:
            return tools[0]
        return None

    @classmethod
    def _get_tool_message(cls, tool: AzureTool, output: str) -> ToolMessage:
        """Returns a tool message for the tool call."""
        tool_message = ToolMessage(
            content=output,
            tool_call_id=tool.tool_call.id,
        )
        tool_message.name = tool._name()  # pyright: ignore [reportCallIssue, reportAttributeAccessIssue]
        return tool_message

    @classmethod
    def tool_message_params(
        cls, tools_and_outputs: list[tuple[AzureTool, str]]
    ) -> list[ToolMessage]:
        """Returns the tool message parameters for tool call results.

        Args:
            tools_and_outputs: The list of tools and their outputs from which the tool
                message parameters should be constructed.

        Returns:
            The list of constructed `ChatCompletionToolMessageParam` parameters.
        """
        return [
            cls._get_tool_message(tool, output) for tool, output in tools_and_outputs
        ]



from collections.abc import Callable

from tenacity import AsyncRetrying, Retrying, retry, stop_after_attempt


def retry_decorator(retries: int | Retrying | AsyncRetrying) -> Callable:
    if isinstance(retries, int):
        if retries == 0:
            return lambda x: x
        return retry(stop=stop_after_attempt(retries))
    return retry(
        sleep=retries.sleep,
        stop=retries.stop,
        wait=retries.wait,
        before=retries.before,
        after=retries.after,
        before_sleep=retries.before_sleep,
        reraise=retries.reraise,
        retry_error_cls=retries.retry_error_cls,
        retry_error_callback=retries.retry_error_callback,
    )



from collections.abc import AsyncGenerator, Generator
from typing import Any

import pytest

from mirascope.core.bedrock._utils._handle_stream import (
    handle_stream,
    handle_stream_async,
)
from mirascope.core.bedrock.call_response_chunk import BedrockCallResponseChunk
from mirascope.core.bedrock.tool import BedrockTool


class MockRecommendBook(BedrockTool):
    def call(self, *args: Any, **kwargs: Any) -> Any: ...

    genre: str
    title: str
    author: str

    @classmethod
    def _name(cls) -> str:
        return "RecommendBook"


class MockAnotherTool(BedrockTool):
    def call(self, *args: Any, **kwargs: Any) -> Any: ...

    @classmethod
    def _name(cls) -> str:
        return "AnotherTool"


@pytest.fixture
def mock_response_text_with_tool():
    return [
        {"messageStart": {"role": "assistant"}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": "Here"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " is"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " a"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " book"}}},
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"text": " recommendation"},
            }
        },
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " for"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " a"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " fantasy"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " book"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": ":"}}},
        {"contentBlockStop": {"contentBlockIndex": 0}},
        {
            "contentBlockStart": {
                "contentBlockIndex": 1,
                "start": {
                    "toolUse": {
                        "name": "RecommendBook",
                        "toolUseId": "tooluse_VQy5Xa-oTXqjpQTf4q6LVg",
                    }
                },
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": ""}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": '{"genre": "f'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": 'antasy"'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": ', "tit'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": 'le": '}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": '"T'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": "he "}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": "Name of"}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": " the"}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": ' Wind"'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": ', "auth'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": 'or": "Patri'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": "ck Rothfu"}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 1,
                "delta": {"toolUse": {"input": 'ss"}'}},
            }
        },
        {"contentBlockStop": {"contentBlockIndex": 1}},
        {"messageStop": {"stopReason": "tool_use"}},
        {
            "metadata": {
                "metrics": {"latencyMs": 1063},
                "usage": {"inputTokens": 395, "outputTokens": 85, "totalTokens": 480},
            }
        },
    ]


@pytest.fixture
def mock_response_only_tool():
    return [
        {"messageStart": {"role": "assistant"}},
        {
            "contentBlockStart": {
                "contentBlockIndex": 0,
                "start": {
                    "toolUse": {
                        "name": "RecommendBook",
                        "toolUseId": "tooluse_Z3qh_1hxSoCzEIw5yBiVpQ",
                    }
                },
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": ""}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": '{"g'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": 'enre": "'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": 'fantasy"'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": ', "title": '}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": '"The Way'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": ' of Kings"'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": ", "}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": '"author": "'}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": "Brandon "}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": "Sand"}},
            }
        },
        {
            "contentBlockDelta": {
                "contentBlockIndex": 0,
                "delta": {"toolUse": {"input": 'erson"}'}},
            }
        },
        {"contentBlockStop": {"contentBlockIndex": 0}},
        {"messageStop": {"stopReason": "tool_use"}},
        {
            "metadata": {
                "metrics": {"latencyMs": 876},
                "usage": {"inputTokens": 385, "outputTokens": 74, "totalTokens": 459},
            }
        },
    ]


@pytest.fixture
def mock_response_without_tool():
    return [
        {"messageStart": {"role": "assistant"}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": "Hello"}}},
        {"contentBlockDelta": {"contentBlockIndex": 0, "delta": {"text": " world"}}},
        {"contentBlockStop": {"contentBlockIndex": 0}},
        {"messageStop": {"stopReason": "end_turn"}},
        {
            "metadata": {
                "metrics": {"latencyMs": 100},
                "usage": {"inputTokens": 10, "outputTokens": 2, "totalTokens": 12},
            }
        },
    ]


def test_handle_stream_with_text_and_tool(mock_response_text_with_tool):
    def mock_stream() -> Generator[dict, None, None]:
        yield from mock_response_text_with_tool

    tool_types = [MockRecommendBook]
    results = list(handle_stream(mock_stream(), tool_types))  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert isinstance(results[-2][1], MockRecommendBook)
    assert results[-2][1].genre == "fantasy"
    assert results[-2][1].title == "The Name of the Wind"
    assert results[-2][1].author == "Patrick Rothfuss"


@pytest.mark.asyncio
async def test_handle_stream_async_with_text_and_tool(mock_response_text_with_tool):
    async def mock_stream() -> AsyncGenerator[dict, None]:
        for chunk in mock_response_text_with_tool:
            yield chunk

    tool_types = [MockRecommendBook]
    results = tuple([c async for c in handle_stream_async(mock_stream(), tool_types)])  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert isinstance(results[-2][1], MockRecommendBook)
    assert results[-2][1].genre == "fantasy"
    assert results[-2][1].title == "The Name of the Wind"
    assert results[-2][1].author == "Patrick Rothfuss"


def test_handle_stream_with_only_tool(mock_response_only_tool):
    def mock_stream() -> Generator[dict, None, None]:
        yield from mock_response_only_tool

    tool_types = [MockRecommendBook]
    results = list(handle_stream(mock_stream(), tool_types))  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert isinstance(results[2][1], MockRecommendBook)
    assert results[2][1].genre == "fantasy"
    assert results[2][1].title == "The Way of Kings"
    assert results[2][1].author == "Brandon Sanderson"


def test_handle_stream_without_tools(mock_response_without_tool):
    def mock_stream() -> Generator[dict, None, None]:
        yield from mock_response_without_tool

    results = list(handle_stream(mock_stream(), None))  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert all(tool is None for chunk, tool in results)


def test_handle_stream_with_mismatched_tool_name(mock_response_text_with_tool):
    def mock_stream() -> Generator[dict, None, None]:
        yield from mock_response_text_with_tool

    tool_types = [MockAnotherTool]
    results = list(handle_stream(mock_stream(), tool_types))  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert all(tool is None for chunk, tool in results)


def test_handle_stream_with_multiple_tool_types(mock_response_text_with_tool):
    def mock_stream() -> Generator[dict, None, None]:
        yield from mock_response_text_with_tool

    tool_types = [MockAnotherTool, MockRecommendBook]
    results = list(handle_stream(mock_stream(), tool_types))  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert isinstance(results[-2][1], MockRecommendBook)
    assert results[-2][1].genre == "fantasy"
    assert results[-2][1].title == "The Name of the Wind"
    assert results[-2][1].author == "Patrick Rothfuss"


@pytest.mark.asyncio
async def test_handle_stream_async_with_mismatched_tool_name(
    mock_response_text_with_tool,
):
    async def mock_stream() -> AsyncGenerator[dict, None]:
        for chunk in mock_response_text_with_tool:
            yield chunk

    tool_types = [MockAnotherTool]
    results = [c async for c in handle_stream_async(mock_stream(), tool_types)]  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert all(tool is None for chunk, tool in results)


@pytest.mark.asyncio
async def test_handle_stream_async_with_multiple_tool_types(
    mock_response_text_with_tool,
):
    async def mock_stream() -> AsyncGenerator[dict, None]:
        for chunk in mock_response_text_with_tool:
            yield chunk

    tool_types = [MockAnotherTool, MockRecommendBook]
    results = [c async for c in handle_stream_async(mock_stream(), tool_types)]  # pyright: ignore [reportArgumentType]

    assert all(isinstance(chunk, BedrockCallResponseChunk) for chunk, tool in results)
    assert isinstance(results[-2][1], MockRecommendBook)
    assert results[-2][1].genre == "fantasy"
    assert results[-2][1].title == "The Name of the Wind"
    assert results[-2][1].author == "Patrick Rothfuss"



from mirascope.core import Messages, anthropic
from pydantic import BaseModel, ConfigDict, Field


class Book(BaseModel):
    title: str = Field(..., examples=["THE NAME OF THE WIND"])
    author: str = Field(..., examples=["Rothfuss, Patrick"])

    model_config = ConfigDict(
        json_schema_extra={
            "examples": [
                {"title": "THE NAME OF THE WIND", "author": "Rothfuss, Patrick"}
            ]
        }
    )


@anthropic.call("claude-3-5-sonnet-20240620", response_model=Book, json_mode=True)
def extract_book(text: str) -> Messages.Type:
    return Messages.User(
        f"Extract {text}. Match example format excluding 'examples' key."
    )


book = extract_book("The Way of Kings by Brandon Sanderson")
print(book)
# Output: title='THE NAME OF THE WIND' author='Rothfuss, Patrick'



from mirascope.core import BaseMessageParam, vertex
from mirascope.integrations.langfuse import with_langfuse


@with_langfuse()
@vertex.call("gemini-1.5-flash")
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


print(recommend_book("fantasy"))



from typing import Annotated

from mirascope.core import BaseDynamicConfig, prompt_template, vertex
from mirascope.retries.tenacity import collect_errors
from pydantic import AfterValidator, ValidationError
from tenacity import retry, stop_after_attempt


def is_upper(v: str) -> str:
    assert v.isupper(), "Must be uppercase"
    return v


@retry(stop=stop_after_attempt(3), after=collect_errors(ValidationError))
@vertex.call(
    "gemini-1.5-flash",
    response_model=Annotated[str, AfterValidator(is_upper)],  # pyright: ignore [reportArgumentType, reportCallIssue]
)
@prompt_template(
    """
    {previous_errors}

    Who wrote {book}?
    """
)
def identify_author(
    book: str, *, errors: list[ValidationError] | None = None
) -> BaseDynamicConfig:
    previous_errors = None
    if errors:
        previous_errors = f"Previous Errors: {errors}"
        print(previous_errors)
    return {"computed_fields": {"previous_errors": previous_errors}}


author = identify_author("The Name of the Wind")
print(author)
# Previous Errors: [1 validation error for str
# value
#   Assertion failed, Must be uppercase [type=assertion_error, input_value='Patrick Rothfuss', input_type=str]
#     For further information visit https://errors.pydantic.dev/2.7/v/assertion_error]
# PATRICK ROTHFUSS



from mirascope.core import Messages, mistral
from tenacity import retry, stop_after_attempt, wait_exponential


@mistral.call("mistral-large-latest", stream=True)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
def stream():
    for chunk, _ in recommend_book("fantasy"):
        print(chunk.content, end="", flush=True)


stream()



import json

from azure.ai.inference import ChatCompletionsClient
from azure.ai.inference.models import (
    ChatCompletionsToolDefinition,
    ChatRequestMessage,
    FunctionDefinition,
)
from azure.core.credentials import AzureKeyCredential

client = ChatCompletionsClient(
    endpoint="YOUR_ENDPOINT", credential=AzureKeyCredential("YOUR_KEY")
)


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


def identify_author(book: str) -> str:
    completion = client.complete(
        model="gpt-4o-mini",
        messages=[
            ChatRequestMessage({"role": "user", "content": f"Who wrote {book}?"})
        ],
        tools=[
            ChatCompletionsToolDefinition(
                function=FunctionDefinition(
                    name="get_book_author",
                    description="Returns the author of the book with the given title.",
                    parameters={
                        "properties": {"title": {"type": "string"}},
                        "required": ["title"],
                        "type": "object",
                    },
                )
            )
        ],
    )
    if tool_calls := completion.choices[0].message.tool_calls:
        if tool_calls[0].function.name == "get_book_author":
            return get_book_author(**json.loads(tool_calls[0].function.arguments))
    return str(completion.choices[0].message.content)


author = identify_author("The Name of the Wind")
print(author)



from mirascope.core import Messages, bedrock


def parse_recommendation(response: bedrock.BedrockCallResponse) -> tuple[str, str]:
    title, author = response.content.split(" by ")
    return (title, author)


@bedrock.call(
    "anthropic.claude-3-haiku-20240307-v1:0", output_parser=parse_recommendation
)
def recommend_book(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book. Output only Title by Author")


print(recommend_book("fantasy"))
# Output: ('"The Name of the Wind"', 'Patrick Rothfuss')



import json

from mirascope.core import prompt_template, vertex


@vertex.call("gemini-1.5-flash", json_mode=True)
@prompt_template("Provide the author and genre of {book_title}")
def get_book_info(book_title: str): ...


try:
    response = get_book_info("The Name of the Wind")
    print(json.loads(response.content))
except json.JSONDecodeError:
    print("The model produced invalid JSON")



from mirascope.core import BaseDynamicConfig, BaseMessageParam, openai


@openai.call(model="gpt-4o-mini")
def summarize(text: str) -> str:
    return f"Summarize this text: {text}"


@openai.call(model="gpt-4o-mini")
def summarize_and_translate(text: str, language: str) -> BaseDynamicConfig:
    summary = summarize(text)
    return {
        "messages": [
            BaseMessageParam(
                role="user",
                content=f"Translate this text to {language}: {summary.content}",
            )
        ],
        "computed_fields": {"summary": summary},
    }


response = summarize_and_translate("Long English text here...", "french")
print(response.content)
print(
    response.model_dump()["computed_fields"]
)  # This will contain the `summarize` response



from mirascope.core import Messages, azure
from pydantic import BaseModel


class Book(BaseModel):
    """An extracted book."""

    title: str
    author: str


@azure.call("gpt-4o-mini", response_model=Book, json_mode=True)
def extract_book(text: str) -> Messages.Type:
    return Messages.User(f"Extract {text}")


book = extract_book("The Name of the Wind by Patrick Rothfuss")
print(book)
# Output: title='The Name of the Wind' author='Patrick Rothfuss'



from mirascope.core import BaseMessageParam, bedrock
from mirascope.integrations.langfuse import with_langfuse


@with_langfuse()
@bedrock.call("anthropic.claude-3-haiku-20240307-v1:0")
def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book.")]


print(recommend_book("fantasy"))



from collections.abc import AsyncGenerator, Awaitable, Callable, Coroutine
from typing import (
    Any,
    Protocol,
    TypeAlias,
    TypeVar,
)

from ..tool import OpenAIRealtimeTool

_ToolTypeBase: TypeAlias = type[OpenAIRealtimeTool] | Callable

_ResponseBaseType = str | bytes
_ResponseT = TypeVar("_ResponseT", bound=_ResponseBaseType)
_ReceiverArgumentT = TypeVar("_ReceiverArgumentT")
_ToolType = TypeVar("_ToolType", bound=_ToolTypeBase)
# TODO: Improve the type of context
Context: TypeAlias = dict[str, Any]
_OpenAIRealtimeToolT = TypeVar("_OpenAIRealtimeToolT", bound=OpenAIRealtimeTool)


class SenderFunc(Protocol[_ResponseT]):
    def __call__(
        self, context: Context
    ) -> (
        AsyncGenerator[
            _ResponseT | tuple[_ResponseT, list[type[OpenAIRealtimeTool] | Callable]],
            None,
        ]
        | Coroutine[
            Any,
            Any,
            _ResponseT | tuple[_ResponseT, list[type[OpenAIRealtimeTool] | Callable]],
        ]
        | Awaitable[
            _ResponseT | tuple[_ResponseT, list[type[OpenAIRealtimeTool] | Callable]]
        ]
    ): ...


class ReceiverFunc(Protocol[_ReceiverArgumentT]):
    def __call__(
        self, response: _ReceiverArgumentT, context: Context
    ) -> Awaitable[None]: ...


class FunctionCallHandlerFunc(Protocol):
    def __call__(
        self, tool: _OpenAIRealtimeToolT, context: Context
    ) -> Awaitable[str] | str: ...



from mirascope.core import Messages, openai
from openai import AsyncOpenAI


@openai.call("gpt-4o-mini", client=AsyncOpenAI())
async def recommend_book_async(genre: str) -> Messages.Type:
    return Messages.User(f"Recommend a {genre} book")



from google.generativeai import GenerativeModel
from mirascope.core import gemini


@gemini.call("", client=GenerativeModel(model_name="gemini-1.5-flash"))
async def recommend_book_async(genre: str) -> str:
    return f"Recommend a {genre} book"



import asyncio

from mirascope.core import BaseMessageParam, azure


@azure.call(model="gpt-4o-mini", stream=True)
async def recommend_book(genre: str) -> list[BaseMessageParam]:
    return [BaseMessageParam(role="user", content=f"Recommend a {genre} book")]


async def main():
    stream = await recommend_book("fantasy")
    async for chunk, _ in stream:
        print(chunk.content, end="", flush=True)


asyncio.run(main())



from anthropic import Anthropic

client = Anthropic()


def get_book_author(title: str) -> str:
    if title == "The Name of the Wind":
        return "Patrick Rothfuss"
    elif title == "Mistborn: The Final Empire":
        return "Brandon Sanderson"
    else:
        return "Unknown"


def identify_author(book: str) -> str:
    message = client.messages.create(
        model="claude-3-5-sonnet-20240620",
        max_tokens=1024,
        messages=[{"role": "user", "content": f"Who wrote {book}?"}],
        tools=[
            {
                "name": "get_book_author",
                "description": "Returns the author of the book with the given title.",
                "input_schema": {
                    "properties": {"title": {"type": "string"}},
                    "required": ["title"],
                    "type": "object",
                },
            }
        ],
    )
    content = ""
    for block in message.content:
        if block.type == "tool_use":
            if block.name == "get_book_author":
                return get_book_author(**block.input)  # pyright: ignore [reportCallIssue]
        else:
            content += block.text
    return content


author = identify_author("The Name of the Wind")
print(author)



from mirascope.core import mistral, prompt_template
from mistralai.client import MistralClient


@mistral.call("mistral-large-latest", client=MistralClient())
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...



from mirascope.core import azure, prompt_template
from tenacity import retry, stop_after_attempt, wait_exponential


@retry(
    stop=stop_after_attempt(3),
    wait=wait_exponential(multiplier=1, min=4, max=10),
)
@azure.call("gpt-4o-mini")
@prompt_template("Recommend a {genre} book")
def recommend_book(genre: str): ...


print(recommend_book("fantasy"))



import inspect
from collections.abc import Awaitable, Callable
from typing import Any, TypeVar, cast, overload

from typing_extensions import TypeIs

from ..dynamic_config import BaseDynamicConfig
from ..message_param import BaseMessageParam

_BaseDynamicConfigT = TypeVar(
    "_BaseDynamicConfigT", bound=BaseDynamicConfig, covariant=True
)


def _is_message_param_list(result: object) -> TypeIs[list[BaseMessageParam]]:
    return isinstance(result, list)


@overload
def get_dynamic_configuration(
    fn: Callable[..., _BaseDynamicConfigT | list[BaseMessageParam]],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> _BaseDynamicConfigT: ...


@overload
def get_dynamic_configuration(
    fn: Callable[
        ..., Awaitable[_BaseDynamicConfigT] | Awaitable[list[BaseMessageParam]]
    ],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> Awaitable[_BaseDynamicConfigT]: ...


def get_dynamic_configuration(
    fn: Callable[
        ...,
        _BaseDynamicConfigT
        | list[BaseMessageParam]
        | Awaitable[_BaseDynamicConfigT]
        | Awaitable[list[BaseMessageParam]],
    ],
    args: tuple[Any, ...],
    kwargs: dict[str, Any],
) -> _BaseDynamicConfigT | Awaitable[_BaseDynamicConfigT]:
    result = (
        fn._original_fn(*args, **kwargs)  # pyright: ignore [reportFunctionMemberAccess]
        if hasattr(fn, "_original_fn")
        else fn(*args, **kwargs)
    )
    if inspect.isawaitable(result):

        async def inner_async() -> _BaseDynamicConfigT:
            async_result = await result
            if _is_message_param_list(async_result):
                return cast(_BaseDynamicConfigT, {"messages": async_result})
            return cast(_BaseDynamicConfigT, async_result)

        return inner_async()

    if _is_message_param_list(result):
        return cast(_BaseDynamicConfigT, {"messages": result})
    return result
